-------------------------------------------------------------------------------
--
-- Text-mode VGA controller for the XESS-300E
--
-- Uses an OPB interface, e.g., for use with the Microblaze soft core
--
-- Stephen A. Edwards
-- sedwards@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity opb_xsb300e_vga is
  
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(31 downto 0) := X"FEFF1000";
    C_HIGHADDR   : std_logic_vector(31 downto 0) := X"FEFF1FFF");

  port (
    OPB_Clk        : in std_logic;
    OPB_Rst        : in std_logic;

    -- OPB signals
    OPB_ABus       : in std_logic_vector (31 downto 0);
    OPB_BE         : in std_logic_vector (3 downto 0);
    OPB_DBus       : in std_logic_vector (31 downto 0);
    OPB_RNW        : in std_logic;
    OPB_select     : in std_logic;
    OPB_seqAddr    : in std_logic;

    VGA_DBus       : out std_logic_vector (31 downto 0);
    VGA_errAck     : out std_logic;
    VGA_retry      : out std_logic;
    VGA_toutSup    : out std_logic;
    VGA_xferAck    : out std_logic;

    Pixel_Clock    : in std_logic;

    -- Video signals
    VIDOUT_CLK     : out std_logic;
    VIDOUT_RED     : out std_logic_vector(9 downto 0);
    VIDOUT_GREEN   : out std_logic_vector(9 downto 0);
    VIDOUT_BLUE    : out std_logic_vector(9 downto 0);
    VIDOUT_BLANK_N : out std_logic;
    VIDOUT_HSYNC_N : out std_logic;
    VIDOUT_VSYNC_N : out std_logic);

end opb_xsb300e_vga;

architecture Behavioral of opb_xsb300e_vga is
  
  constant BASEADDR : std_logic_vector(31 downto 0) := X"FEFF1000";

  -- Video parameters
  
  constant HTOTAL : integer := 800;
  constant HSYNC : integer := 96;
  constant HBACK_PORCH : integer := 48;
  constant HACTIVE : integer := 640;
  constant HFRONT_PORCH : integer := 16;
  
  constant VTOTAL : integer := 525;
  constant VSYNC : integer := 2;
  constant VBACK_PORCH : integer := 33;
  constant VACTIVE : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  -- 512 X 8 dual-ported Xilinx block RAM
  component RAMB4_S8_S8
    port (
      DOA   : out std_logic_vector (7 downto 0);
      ADDRA : in std_logic_vector (8 downto 0);
      CLKA  : in std_logic;
      DIA   : in std_logic_vector (7 downto 0);
      ENA   : in std_logic;
      RSTA  : in std_logic;
      WEA   : in std_logic;
      DOB   : out std_logic_vector (7 downto 0);
      ADDRB : in std_logic_vector (8 downto 0);
      CLKB  : in std_logic;
      DIB   : in std_logic_vector (7 downto 0);
      ENB   : in std_logic;
      RSTB  : in std_logic;
      WEB   : in std_logic);
  end component;

  -- Signals latched from the OPB
  signal ABus : std_logic_vector (31 downto 0);
  signal DBus : std_logic_vector (31 downto 0);
  signal RNW  : std_logic;
  signal select_delayed : std_logic;

  -- Latched output signals for the OPB
  signal DBus_out : std_logic_vector (31 downto 0);

  -- Signals for the OPB-mapped RAM
  signal ChipSelect : std_logic;            -- Address decode
  signal MemCycle1, MemCycle2 : std_logic;  -- State bits
  signal RamPageAddress : std_logic_vector(2 downto 0); 
  signal RamSelect : std_logic_vector (7 downto 0);
  signal RST, WE : std_logic_vector (7 downto 0);
  signal DOUT0, DOUT1, DOUT2, DOUT3 : std_logic_vector(7 downto 0);
  signal DOUT4, DOUT5, DOUT6, DOUT7 : std_logic_vector(7 downto 0);
  signal ReadData : std_logic_vector(7 downto 0);

  -- Signals for the video controller
  signal LoadNShift : std_logic;        -- Shift register control
  signal FontData  : std_logic_vector(7 downto 0);  -- Input to shift register
  signal ShiftData : std_logic_vector(7 downto 0);  -- Shift register data
  signal VideoData : std_logic;         -- Serial out ANDed with blanking

  signal Hcount : std_logic_vector(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : std_logic_vector(9 downto 0);  -- Vertical position (0-524)
  signal HBLANK_N, VBLANK_N : std_logic;   -- Blanking signals

  signal START_POSITION : std_logic_vector(9 downto 0);
  signal STOP_POSITION : std_logic_vector(9 downto 0);

  signal START_POSITIONa : std_logic_vector(7 downto 0);
  signal STOP_POSITIONa : std_logic_vector(7 downto 0);  

  
  signal COLOR : std_logic_vector(7 downto 0);

  
--  signal FontLoad, LoadChar : std_logic;  -- Font/Character RAM read triggers
--signal FontAddr : std_logic_vector(10 downto 0);
--  signal CharRamPage : std_logic_vector(2 downto 0);
--  signal CharRamSelect_N : std_logic_vector(4 downto 0);
--  signal FontRamPage : std_logic_vector(1 downto 0);
--  signal FontRamSelect_N : std_logic_vector(2 downto 0);
--  signal CharAddr : std_logic_vector(11 downto 0);
--  signal CharColumn : std_logic_vector(9 downto 0);
--  signal CharRow : std_logic_vector(9 downto 0);
--  signal Column : std_logic_vector(6 downto 0); -- 0-79
--  signal Row : std_logic_vector(4 downto 0);   -- 0-29
  signal EndOfLine, EndOfField : std_logic;

  signal DOUTB0, DOUTB1, DOUTB2, DOUTB3 : std_logic_vector(7 downto 0);
  signal DOUTB4, DOUTB5, DOUTB6, DOUTB7 : std_logic_vector(7 downto 0);

  signal msel : std_logic;
  
begin  -- Behavioral

  -- Port A is used for communication with the OPB
  -- Port B is for video

process(pixel_clock)
  begin
    if pixel_clock'event and pixel_clock = '1' then
      msel <= Hcount(9);
    end if;
  end process;  
  
  START_POSITIONa <= DOUTB1 when (msel = '1') else DOUTB0;
  STOP_POSITIONa  <= DOUTB3 when (msel = '1') else DOUTB2;


  START_POSITION <= "00" & START_POSITIONa;
  STOP_POSITION <= ("00" & STOP_POSITIONa) + 240;
  
  COLOR          <= DOUTB5 when (msel='1') else DOUTB4;


  
  RAMB4_S8_S8_0 : RAMB4_S8_S8
    port map (
      DOA   => DOUT0,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(0),
      WEA   => WE(0),
      DOB   => DOUTB0,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

  RAMB4_S8_S8_1 : RAMB4_S8_S8
    port map (
      DOA   => DOUT1,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(1),
      WEA   => WE(1),
      DOB   => DOUTB1,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

    RAMB4_S8_S8_2 : RAMB4_S8_S8
    port map (
      DOA   => DOUT2,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(2),
      WEA   => WE(2),
      DOB   => DOUTB2,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

    RAMB4_S8_S8_3 : RAMB4_S8_S8
    port map (
      DOA   => DOUT3,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(3),
      WEA   => WE(3),
      DOB   => DOUTB3,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

    RAMB4_S8_S8_4 : RAMB4_S8_S8
    port map (
      DOA   => DOUT4,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(4),
      WEA   => WE(4),
      DOB   => DOUTB4,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

    RAMB4_S8_S8_5 : RAMB4_S8_S8
    port map (
      DOA   => DOUT5,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(5),
      WEA   => WE(5),
      DOB   => DOUTB5,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

  RAMB4_S8_S8_6 : RAMB4_S8_S8
    port map (
      DOA   => DOUT6,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(6),
      WEA   => WE(6),
      DOB   => DOUTB6,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

    RAMB4_S8_S8_7 : RAMB4_S8_S8
    port map (
      DOA   => DOUT7,
      ADDRA => ABus(8 downto 0),
      CLKA  => OPB_Clk,
      DIA   => DBus(7 downto 0),
      ENA   => '1',
      RSTA  => RST(7),
      WEA   => WE(7),
      DOB   => DOUTB7,
      ADDRB => Hcount(8 downto 0),
      CLKB  => Pixel_Clock,
      DIB   => X"00",
      ENB   => '1',
      RSTB  => '0',
      WEB   => '0');

  -----------------------------------------------------------------------------
  --  
  -- OPB-RAM controller
  --
  -----------------------------------------------------------------------------
  
  -- Unused OPB control signals
  VGA_errAck <= '0';
  VGA_retry <= '0';
  VGA_toutSup <= '0';

  -- Latch the relevant OPB signals from the OPB, since they arrive late
  LatchOPB: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Abus <= ( others => '0' );
      DBus <= ( others => '0' );
      RNW <= '1';
      select_delayed <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      ABus <= OPB_ABus;
      DBus <= OPB_DBus;
      RNW <= OPB_RNW;
      select_delayed <= OPB_Select;
    end if;
  end process LatchOPB;

  -- Address bits 31 downto 12 is our chip select
  -- 11 downto 9 is the RAM page select
  -- 8 downto 0 is the RAM byte select
  
  ChipSelect <=
    '1' when select_delayed = '1' and
       (ABus(31 downto 12) = BASEADDR(31 downto 12)) and
       MemCycle1 = '0' and MemCycle2 = '0' else
    '0';

  RamPageAddress <= ABus(11 downto 9);
  
  RamSelect <=
    "00000001" when RamPageAddress = "000" else
    "00000010" when RamPageAddress = "001" else
    "00000100" when RamPageAddress = "010" else
    "00001000" when RamPageAddress = "011" else
    "00010000" when RamPageAddress = "100" else
    "00100000" when RamPageAddress = "101" else
    "01000000" when RamPageAddress = "110" else
    "10000000" when RamPageAddress = "111" else
    "00000000";

  MemCycleFSM : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      MemCycle1 <= '0';
      MemCycle2 <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      MemCycle2 <= MemCycle1;
      MemCycle1 <= ChipSelect;
    end if;
  end process MemCycleFSM;
  
  VGA_xferAck <= MemCycle2;

  WE <=
  RamSelect when ChipSelect = '1' and RNW = '0' and OPB_Rst = '0' else
    "00000000";
 
  RST <=
   not RamSelect when ChipSelect = '1' and RNW = '1' and OPB_Rst = '0' else
    "11111111"; 

  ReadData <= DOUT0 or DOUT1 or DOUT2 or DOUT3 or
              DOUT4 or DOUT5 or DOUT6 or DOUT7 when MemCycle1 = '1'
              else "00000000";

  -- DBus(31 downto 24) is the byte for addresses ending in 0

  GenDOut: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      DBus_out <= ( others => '0');
    elsif OPB_Clk'event and OPB_Clk = '1' then
      DBus_out <= ReadData & ReadData & ReadData & ReadData;
    end if;
  end process GenDOut;

  VGA_DBus <= DBus_out;

  ----------------------------------------------------------------------------
  --
  -- Video controller
  --
  -----------------------------------------------------------------------------

  -- Horizontal and vertical counters

  HCounter : process (Pixel_Clock, OPB_Rst)
  begin    
    if OPB_Rst = '1' then
      Hcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;      
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';
 
  VCounter: process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Vcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (Pixel_Clock, OPB_Rst)
  begin    
    if OPB_Rst = '1' then
      VIDOUT_HSYNC_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        VIDOUT_HSYNC_N <= '0';
      elsif Hcount = HSYNC - 1 then
        VIDOUT_HSYNC_N <= '1';
      end if;      
    end if;
  end process HSyncGen;

  -- The -1 correction doesn't appear here to correct for the
  -- registered video signal outputs.
  
  HBlankGen : process (Pixel_Clock, OPB_Rst)
  begin    
    if OPB_Rst = '1' then
      HBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if Hcount = HSYNC + HBACK_PORCH then
        HBLANK_N <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        HBLANK_N <= '0';
      end if;      
    end if;
  end process HBlankGen;

  VSyncGen : process (Pixel_Clock, OPB_Rst)
  begin    
    if OPB_Rst = '1' then
      VIDOUT_VSYNC_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if EndOfField = '1' then
          VIDOUT_VSYNC_N <= '0';
        elsif VCount = VSYNC - 1 then
          VIDOUT_VSYNC_N <= '1';
        end if;
      end if;      
    end if;
  end process VSyncGen;

  VBlankGen : process (Pixel_Clock, OPB_Rst)
  begin    
    if OPB_Rst = '1' then
      VBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          VBLANK_N <= '1';
        elsif VCount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          VBLANK_N <= '0';
        end if;
      end if;      
    end if;
  end process VBlankGen;

  -- RAM read triggers and shift register control

--  LoadChar   <= '1' when Hcount(2 downto 0) = X"5" else '0';
--  FontLoad   <= '1' when Hcount(2 downto 0) = X"6" else '0'; 
--  LoadNShift <= '1' when Hcount(2 downto 0) = X"7" else '0';

  -- Correction of 4 needed to calculate the character address before the
  -- character is displayed
--  CharColumn <= Hcount - HSYNC - HBACK_PORCH + 4;
--  Column <= CharColumn(9 downto 3);  -- /8
--  CharRow <= Vcount - VSYNC - VBACK_PORCH;
--  Row <= CharRow(8 downto 4);  -- / 16

  -- Column + Row * 80
--  CharAddr <= Column +
--              ("0" & Row(4 downto 0) & "000000") +
--              ("000" & Row(4 downto 0) & "0000");

--  CharRamPage <= CharAddr(11 downto 9);
--  CharRamSelect_N <=
--    "11110" when CharRamPage = "000" else
--    "11101" when CharRamPage = "001" else
--    "11011" when CharRamPage = "010" else
--    "10111" when CharRamPage = "011" else
--    "01111" when CharRamPage = "100" else
--    "11111";

  -- Most significant bit of character ignored
--  FontAddr(10 downto 4) <=
--    (DOUTB0(6 downto 0) or DOUTB1(6 downto 0) or DOUTB2(6 downto 0) or
--     DOUTB3(6 downto 0) or DOUTB4(6 downto 0));
--  FontAddr(3 downto 0) <= CharRow(3 downto 0);

  -- Unusual addressing: font only holds 96 of 128 possible characters
  -- First 32 characters appear twice
--  FontRamPage <= FontAddr(10 downto 9);
--  FontRamSelect_N <=
--    "110" when FontRamPage = "00" else
--    "110" when FontRamPage = "01" else
--    "101" when FontRamPage = "10" else
--    "011" when FontRamPage = "11" else
--    "111";

--  FontData <= DOUTB5 or DOUTB6 or DOUTB7;
  
  -- Shift register
--  ShiftRegister: process (Pixel_Clock, OPB_Rst)
--  begin
--    if OPB_Rst = '1' then
--      ShiftData <= X"AB";
--    elsif Pixel_Clock'event and Pixel_Clock = '1' then
--     if LoadNShift = '1' then
--        ShiftData <= FontData;
--      else
--        ShiftData <= ShiftData(6 downto 0) & ShiftData(7);
--      end if;
--    end if;
--  end process ShiftRegister;
--  VideoData <= ShiftData(7);

  -- Registered video signals going to the video DAC
  VideoOut: process (Pixel_Clock, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      VIDOUT_BLANK_N <= '0';
      VIDOUT_RED   <= "0000000000";
      VIDOUT_BLUE  <= "0000000000";
      VIDOUT_GREEN <= "0000000000";
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      VIDOUT_BLANK_N <= VBLANK_N and HBLANK_N;

      
      --color pick logic here
      if (Vcount < START_POSITION) then
        VIDOUT_RED   <= "1000000000";
        VIDOUT_GREEN <= "1000000000";  
        VIDOUT_BLUE  <= "1111111111";
      elsif (Vcount > STOP_POSITION) then  --240
        VIDOUT_RED   <= "1100000000";
        VIDOUT_GREEN <= "1100000000";
        VIDOUT_BLUE  <= "1100000000";
      else
        VIDOUT_RED   <= COLOR(7 downto 5) & "0000000";
        VIDOUT_GREEN <= COLOR(4 downto 2) & "0000000";
        VIDOUT_BLUE  <= COLOR(1 downto 0) & "00000000";
      end if;
    end if;
  end process VideoOut;

  VIDOUT_CLK <= Pixel_Clock;
end Behavioral;
